//Master Lenman's XOR Encryption 2.0 - A Polecatsoft program
//This program encrypts and decrypts files using an extremely simple algorithm
//Same program for encryption and decryption

#include <string>
#include <iostream>
#include <fstream>

#include "treeNode.h"
#include "treeStringClass.h"
#include "MLTCClass.h"
#include "main.h"

using namespace std;

MLTCStream::MLTCStream() : DEFAULT_SIZE(8)
{
	//Initialise some variables
	iSize = oSize = DEFAULT_SIZE;
	iBuffer = new char[DEFAULT_SIZE];
	oBuffer = new char[DEFAULT_SIZE];
	iUsed = oUsed = 0;
	iPos = oPos = 0;
	curOutputBitPos = 0;
	tempBitByte = 0;
	curInputBitPos = 0;
}

//Destructor
MLTCStream::~MLTCStream()
{
	delete [] iBuffer;
	delete [] oBuffer;
}

//Doubles the size of the input buffer
void MLTCStream::resizeInputBuffer()
{
	//iSize = iSize * 2;
	iSize <<= 1;
	//Create temp buffer with new size
	char *temp = new char[iSize];
	//Copy data
	for(int i = 0; i < iUsed; ++i)
	{
		temp[i] = iBuffer[i];
	}
	//Destroy old buffer
	delete [] iBuffer;
	iBuffer = temp;
}

//Doubles the size of the output buffer
void MLTCStream::resizeOutputBuffer()
{
	//oSize = oSize * 2;
	oSize <<= 1;
	//Create temp buffer with new size
	char *temp = new char[oSize];
	//Copy data
	for(int i = 0; i < oUsed; ++i)
	{
		temp[i] = oBuffer[i];
	}
	//Destroy old buffer
	delete [] oBuffer;
	oBuffer = temp;
}

//Insert input byte
void MLTCStream::put(char byte)
{
	if(iUsed == iSize)
	{
		resizeInputBuffer();
	}
	iBuffer[iUsed++] = byte;
}

//Get output byte
char MLTCStream::get()
{
	if(oPos >= oUsed)
		throw;
	return oBuffer[oPos++];
}

//Write n bytes to the input buffer
void MLTCStream::write(const char *ins, int n)
{
	while(iUsed + n > iSize)
	{
		resizeInputBuffer();
	}
	for(int i = 0; i < n; ++i)
	{
		iBuffer[iUsed++] = ins[i];
	}
}

//Write n bytes from the output buffer to arr
void MLTCStream::read(char *arr, int n)
{
	if(oPos + n > oUsed)
	{
		throw;
	}
	for(int i = 0; i < n; ++i)
	{
		arr[i] = oBuffer[oPos++];
	}
}


//Get one byte from the input buffer
char MLTCStream::inputGet()
{
	return iBuffer[iPos++];
}

//Read n bytes from the input buffer
void MLTCStream::readInput(char *bytes, int n)
{
	if(iPos + n > iUsed)
	{
		throw;
	}
	for(int i = 0; i < n; ++i)
	{
		bytes[i] = iBuffer[iPos++];
	}
}

//Clear input buffer
void MLTCStream::clearInputBuffer()
{
	iUsed = iPos = 0;
	iSize = DEFAULT_SIZE;
	delete [] iBuffer;
	iBuffer = new char[DEFAULT_SIZE];
}

//Clear output buffer
void MLTCStream::clearOutputBuffer()
{
	oUsed = oPos = 0;
	oSize = DEFAULT_SIZE;
	delete [] oBuffer;
	oBuffer = new char[DEFAULT_SIZE];
	curOutputBitPos = 0;
}

//Clear all buffers
void MLTCStream::clear()
{
	clearInputBuffer();
	clearOutputBuffer();
}

//Compress input
void MLTCStream::compress(bool showProgress)
{
	//Look how often each character is used
	treeNode nodes[511];
	for(int i = 0; i < iUsed; ++(nodes[unsigned char(iBuffer[i++])].frequency));

	//Create the binary tree
	//These pos variables store the position of the smallest value
	int pos1, pos2;
	//The currently used parent node index
	int pNodeIndex = 256;
	int highestNodeIndex = -1;
	while(true)
	{
		pos1 = pos2 = -1;
		for(int i = 0; i < 511; ++i)
		{
			//Check if the node can still be used
			if(nodes[i].parent || !nodes[i].frequency)
				continue;
			if(pos1 < 0)
			{
				pos1 = i;
			}
			else if(pos2 < 0)
			{
				pos2 = i;
			}
			else
			{
				if(nodes[i].frequency < nodes[pos1].frequency || nodes[i].frequency < nodes[pos2].frequency)
				{
					if(nodes[pos1].frequency > nodes[pos2].frequency)
						pos1 = i;
					else
						pos2 = i;
				}
			}
		}
		//If only the first position is found, we know that this is the highest node
		if(pos2 < 0)
		{
			highestNodeIndex = pos1;
			break;
		}
		nodes[pos1].parent = nodes[pos2].parent = &(nodes[pNodeIndex]);
		nodes[pNodeIndex].child1 = &(nodes[pos1]);
		nodes[pNodeIndex].child2 = &(nodes[pos2]);
		nodes[pNodeIndex].frequency = nodes[pos1].frequency + nodes[pos2].frequency;
		++pNodeIndex;
	}
	calculateCodes(nodes);
	//Create tree creation string
	int highestCodeLength = 0;
	for(int i = 0; i < 256; ++i)
	{
		if(nodes[i].codeLength > highestCodeLength)
			highestCodeLength = nodes[i].codeLength;
	}
	//Create output
	writeOutput((char*)(&iUsed), sizeof(iUsed));
	string s;
	int length = 0;
	for(int i = 1; i <= highestCodeLength; ++i)
	{
		char numberOfNodes = 0;
		for(int j = 0; j < 256; ++j)
		{
			if(nodes[j].codeLength == i)
				++numberOfNodes;
		}
		s.append(1, numberOfNodes);
		++length;
		for(int j = 0; j < 256; ++j)
		{
			if(nodes[j].codeLength == i)
			{
				++length;
				s.append(1, j);
			}
		}
	}
	short sLength = short(s.length());
	writeOutput((char*)(&sLength), sizeof(sLength));
	writeOutput(s.c_str(), sLength);

	//Recreate tree based on the creation string
	treeString treeCreation(s.c_str(), nodes, highestCodeLength);


	//Calculate average bits per byte before the nodes are reset
	int bits = 0;
	for(int i = 0; i < 256; ++i)
	{
		bits += nodes[i].frequency * nodes[i].codeLength;
	}
	bitsPerByte = float(bits) / float(iUsed);

	//Reset all nodes
	for(int i = 0; i < 511; ++i)
	{
		nodes[i].reset();
	}

	pNodeIndex = 256;
	assignChildren(pNodeIndex, treeCreation, 1, nodes, highestCodeLength);

	calculateCodes(nodes);

	//For debugging
	/*ofstream debug("debugC.txt");
	debug << "ASCII chart:" <<endl;
	for(int i = 0; i < 256; ++i)
	{
		debug << char(i) << ": code ";byteAsBin(nodes[i].code, &debug);debug<<", length: " << nodes[i].codeLength << endl;
	}
	debug << endl << s;
	debug.close(); */
	//End of debugging

	int stepCount = iUsed / 100;
	//Make sure the percentage never goes OVER 100%, at the cost of it sometimes not
	//reaching the 100% at all
	while(stepCount * 100 < iUsed)
		++stepCount;
	const int STEPSIZE = stepCount;
	int percentage = 0;
	if(showProgress)
		cout << "00%";
	//Create binary output
	for(int i = 0; i < iUsed; ++i)
	{
		outputPutBits(nodes[unsigned char(iBuffer[i])].code, nodes[unsigned char(iBuffer[i])].codeLength);
		if(showProgress)
		{
			if(!(--stepCount))
			{
				stepCount = STEPSIZE;
				cout << char(0x08) << char(0x08) << char(0x08);
				if(++percentage < 10)
					cout << "0";
				cout << percentage << "%";
			}
		}
	}
	flushOutbutBits();

	//Remove the progress characters
	if(showProgress)
	{
		if(percentage >= 100)
			cout << char(0x08);
		cout << char(0x08) << char(0x08) << char(0x08);
	}
}

void MLTCStream::calculateCodes(treeNode *nodes)
{
	for(int i = 0; i < 256; ++i)
	{
		treeNode *curNode = &nodes[i];
		while(curNode->parent)
		{
			//Create next part of code
			if(curNode == curNode->parent->child1)
				nodes[i].code |= 1 << nodes[i].codeLength;

			++nodes[i].codeLength;
			curNode = curNode->parent;
		}
	}
}

int globstuff = 0;

//Recursive testFunction for creating the binary tree
void MLTCStream::assignChildren(int &curPos, treeString &tree, const int depth, treeNode *nodes, const int maxDepth)
{
	const int thisPos = curPos;
	//assign child 1
	treeNode *child;
	if((child = tree.getChildAt(depth)) != 0)
	{
		nodes[thisPos].child1 = child;
		child->parent = &(nodes[thisPos]);
	}
	else
	{
		++curPos;
		nodes[thisPos].child1 = &(nodes[curPos]);
		nodes[curPos].parent = &(nodes[thisPos]);
		assignChildren(curPos, tree, depth + 1, nodes, maxDepth);
	}
	//assign child 2
	if((child = tree.getChildAt(depth)) != 0)
	{
		nodes[thisPos].child2 = child;
		child->parent = &(nodes[thisPos]);
	}
	else
	{
		++curPos;
		nodes[thisPos].child2 = &(nodes[curPos]);
		nodes[curPos].parent = &(nodes[thisPos]);
		assignChildren(curPos, tree, depth + 1, nodes, maxDepth);
	}
}

//Decompress input
void MLTCStream::decompress(bool showProgress)
{
	int decompressedSize;
	readInput((char*)(&decompressedSize), sizeof(decompressedSize));
	short treeStringLength;
	readInput((char*)(&treeStringLength), sizeof(treeStringLength));
	char *treeCharString = new char[treeStringLength + 1];
	treeCharString[treeStringLength] = 0;
	readInput(treeCharString, treeStringLength);

	treeNode nodes[511];
	int pNodeIndex = 256, parentNode = 256;
	for(int i = 0; i < 256; ++i)
	{
		nodes[i].ASCII = i;
	}

	int highestCodeLength = 0;
	for(int i = 0; i < treeStringLength; ++i)
	{
		i += treeCharString[i];
		++highestCodeLength;
	}

	treeString treeCreation(treeCharString, nodes, highestCodeLength);

	assignChildren(pNodeIndex, treeCreation, 1, nodes, highestCodeLength);

	int stepCount = decompressedSize / 100;
	while(stepCount * 100 < decompressedSize)
		++stepCount;
	const int STEPSIZE = stepCount;
	int percentage = 0;
	cout << "00%";
	treeNode *parentNodePointer = &(nodes[parentNode]);
	for(int i = 0; i < decompressedSize; ++i)
	{
		treeNode *curNode = parentNodePointer;
		while(curNode->child1)
		{
			if(inputGetBit())
			{
				curNode = curNode->child1;
			}
			else
			{
				curNode = curNode->child2;
			}
		}
		outputPut(curNode->ASCII);
		if(showProgress)
		{
			if(!(--stepCount))
			{
				stepCount = STEPSIZE;
				cout << char(0x08) << char(0x08) << char(0x08);
				if(++percentage < 10)
					cout << "0";
				cout << percentage << "%";
			}
		}
	}
	if(showProgress)
	{
		if(percentage >= 100)
			cout << char(0x08);
		cout << char(0x08) << char(0x08) << char(0x08);
	}

	//Debugging only, can be removed
	/*calculateCodes(nodes);
	ofstream debug("debugD.txt");
	debug << "ASCII chart:" <<endl;
	for(int i = 0; i < 256; ++i)
	{
		debug << char(i) << ": code ";byteAsBin(nodes[i].code, &debug);debug<<", length: " << nodes[i].codeLength << endl;
	}
	debug << endl << treeCharString;
	debug.close();*/
	//End of debugging

	delete [] treeCharString;
}

//Some getters:

int MLTCStream::getOutputSize()
{
	return oUsed;
}

const char *MLTCStream::getOutputBuffer()
{
	return oBuffer;
}

void MLTCStream::flushOutbutBits()
{
	outputPut(tempBitByte);
	curOutputBitPos = 0;
	tempBitByte = 0;
}

void MLTCStream::outputPut(char byte)
{
	if(oUsed == oSize)
	{
		resizeOutputBuffer();
	}
	oBuffer[oUsed++] = byte;
}

//Write n bytes to the output buffer
void MLTCStream::writeOutput(const char *ins, int n)
{
	while(oUsed + n > oSize)
	{
		resizeOutputBuffer();
	}
	for(int i = 0; i < n; ++i)
	{
		oBuffer[oUsed++] = ins[i];
	}
}

void MLTCStream::outputPutBits(int code, int bits)
{
	for(int i = 0; i < bits; ++i)
	{
		if(curOutputBitPos == 8)
		{
			flushOutbutBits();
		}
		tempBitByte |= ((code >> (bits - i - 1)) & 1) << (7 - curOutputBitPos++);
	}
}

bool MLTCStream::inputGetBit()
{
	bool returnValue = false;
	if(curInputBitPos == 8)
	{
		++iPos;
		curInputBitPos = 0;
	}
	if(iBuffer[iPos] & (1 << (7 - curInputBitPos++)))
		returnValue = true;

	return returnValue;
}

//This will calculate the average bits per byte value
float MLTCStream::getBitsPerByte()
{
	return bitsPerByte;
}

/*
.ct file format specification:
ct stands for Compressed Text. The MLTCStream class creates valid ct data,
using the compress method.
-4 byte, uint, Number of bytes of the encrypted file (see last point)
-2 byte, short, length of the char array (see next point)
-x byte, char array, the binary tree is created out of this string
-x byte, binary data, the compressed data
*/