//Stegano TGA - A Polecatsoft program
//The program will hide the file in parameter 3 in the bitmap specified in parameter 1,
//storing the result in parameter 2.
//The file can be retrieved using Stegano TGA Retrieve

#include <string>
#include <iostream>
#include <fstream>

#include "main.h"
#include "XorEncryption.h"

using namespace std;

//For 16 bit images:
//This function hides the writeByte byte in 6 bytes of input, storing the result in output
void transfer16Byte1(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption)
{
	if(encryption)
		encryption->exe(writeByte);
	unsigned short readData;
	//Hide the first 6 bits from writeByte
	for(int i = 0; i < 6; i += 3)
	{
		input.read(reinterpret_cast<char*>(&readData), 2);
		//This instruction performs all the code below in one step
		readData = (readData & 0xFBDE) | ((writeByte & (1 << i)) << (10 - i))
			                           | ((writeByte & (1 << (i + 1))) << (4 - i)) // == 5 - (i + 1)
									   | ((writeByte >> (i + 2)) & 0x0001);
									   //| ((writeByte & (1 << (i + 2))) >> (i + 2));
#if 0
		//Delete 3 bits from the read data
		readData &= 0xFBDE; //  A RRRRR GGGGG BBBBB
		                    //& 1 11110 11110 11110
		//Get and set first bit
		unsigned char curBit = (writeByte & (1 << i)) >> i;
		readData |= curBit << 10;
		//Get and set second bit
		curBit = (writeByte & (1 << (i + 1))) >> (i + 1);
		readData |= curBit << 5;
		//Get and set third bit
		curBit = (writeByte & (1 << (i + 2))) >> (i + 2);
		readData |= curBit;
#endif
		output.write(reinterpret_cast<char*>(&readData), 2);
	}
	//Hide the last 2 bits
	input.read(reinterpret_cast<char*>(&readData), 2);
	//Delete 2 bits from the read data and set the 2 bits from the writeByte
	//  A RRRRR GGGGG BBBBB
	//& 1 11110 11110 11111
	readData = (readData & 0xFBDF) | ((writeByte & 0x40) << 4) | ((writeByte & 0x80) >> 2);
	output.write(reinterpret_cast<char*>(&readData), 2);
}

//This function hides the writeByte byte in 4 bytes of input, storing the result in output
void transfer16Byte2(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption)
{
	if(encryption)
		encryption->exe(writeByte);
	//Hide the first 6 bits from writeByte
	unsigned short readData;
	input.read(reinterpret_cast<char*>(&readData), 2);
	//This instruction performs all the code below in one step
	readData = (readData & 0xF39C) | ((writeByte & 0x03) << 10)
		                           | ((writeByte & 0x0C) << 3)
		                           | ((writeByte & 0x30) >> 4);
#if 0
	//Delete 6 bits from the read data
	readData &= 0xF39C; //  A RRRRR GGGGG BBBBB
	                    //& 1 11100 11100 11100
	//Get and set first two bits
	unsigned char curBit = writeByte & 0x03;
	readData |= curBit << 10;
	//Get and set second two bits
	curBit = (writeByte & 0x0C) >> 2;
	readData |= curBit << 5;
	//Get and set third two bits
	curBit = (writeByte & 0x30) >> 4;
	readData |= curBit;
#endif
	output.write(reinterpret_cast<char*>(&readData), 2);

	//Hide the last 2 bits
	input.read(reinterpret_cast<char*>(&readData), 2);
	readData = (readData & 0xFBDF) | ((writeByte & 0x40) << 4) | ((writeByte & 0x80) >> 2);
#if 0
	//Delete 2 bits from the read data
	readData &= 0xFBDF; //  A RRRRR GGGGG BBBBB
	                    //& 1 11110 11110 11111
	//Set the 2 bits from the writeByte
	readData |= ((writeByte & 0x40) << 4) | ((writeByte & 0x80) >> 2);
#endif
	output.write(reinterpret_cast<char*>(&readData), 2);
}

//For 24 and 32 bit images:
//This function hides the writeByte byte in 8 bytes of input, storing the result in output
void transferByte1(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption)
{
	if(encryption)
		encryption->exe(writeByte);
	unsigned char readByte;
	for(int i = 0; i < 8; ++i)
	{
		//Get value of active bit
		input.read(reinterpret_cast<char *>(&readByte), 1);
		//Set last bit to new value
		readByte = (readByte & 0xFE) | ((writeByte >> i) & 1);
		output.write(reinterpret_cast<char *>(&readByte), 1);
	}
}

//This function hides the writeByte byte in 4 bytes of input, storing the result in output
void transferByte2(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption)
{
	if(encryption)
		encryption->exe(writeByte);
	for(int i = 0; i < 8; i += 2)
	{
		unsigned char readByte;
		input.read(reinterpret_cast<char *>(&readByte), 1);
		//Set last 2 bits to new value
		readByte = (readByte & 0xFC) | ((writeByte >> i) & 0x03);
		output.write(reinterpret_cast<char *>(&readByte), 1);
	}
}

//This function hides the writeByte byte in 2 bytes of input, storing the result in output
void transferByte4(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption)
{
	if(encryption)
		encryption->exe(writeByte);
	for(int i = 0; i < 8; i += 4)
	{
		unsigned char readByte;
		input.read(reinterpret_cast<char *>(&readByte), 1);
		//Set last 4 bits to new value
		readByte = (readByte & 0xF0) | ((writeByte >> i) & 0x0F);
		output.write(reinterpret_cast<char *>(&readByte), 1);
	}
}

int steganoTgaHide(int argc, char **argv)
{
	//for() counter variable
	int i;

	//Check if enough parameters are given
	cout << "You are now running the Stegano TGA function of MLSS.\n" << endl;
	if(argc < 3)
	{
		cout << "Possible options (explanations are in brackets):\n"
			<< "-h hostfile (TGA file which will serve as host)\n"
			<< "-o outputfile (TGA file that will be created; default: output.tga)\n"
			<< "-g guestfile (file that will be hidden)\n"
			<< "-p password (a password that will be used to encrypt the guestfile)\n"
			<< "-s (scramble unused parts of the file using random data)\n"
			<< "-b bitsperbyte (defines how many bits of each byte of the hostfile will be used to hide data. Possible values are 1, 2 & 4 (only 1 & 2 for 16 bit images);\ndefault: 1)\n";

		cout << "Non-optional options: -h -g" << endl;
		return 0;
	}

	//File names
	string hostfile, outputfile, guestfile, password;
	bool scramble = false;
	encryptClass *encryption = 0;
	//How many bits of each hostfile byte will be used to hide data
	int bitsperbyte = 1;
	void (*transferByte)(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption) = transferByte1;
	for(i = 2; i < argc; ++i)
	{
		string temp(argv[i]);
		bool missingArgument = false;
		if(temp == string("-h"))
		{
			if(++i < argc)
			{
				hostfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-o"))
		{
			if(++i < argc)
			{
				outputfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-g"))
		{
			if(++i < argc)
			{
				guestfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-b"))
		{
			if(++i < argc)
			{
				bitsperbyte = str2int(argv[i]);
				if(bitsperbyte == 1)
					transferByte = transferByte1;
				else if(bitsperbyte == 2)
					transferByte = transferByte2;
				else if(bitsperbyte == 4)
					transferByte = transferByte4;
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-p"))
		{
			if(++i < argc)
			{
				encryption = new encryptClass(argv[i]);
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-s"))
		{
			scramble = true;
			continue;
		}
		if(missingArgument)
			cout << "The option " << temp << " is missing an argument!";
		else
			cout << "Unknown option \"" << temp << "\"!";
		cout << endl;
		return 0;
	}

	//hostfile bytes needed for each guestfile byte
	int bytesperbyte = 8 / bitsperbyte;

	//Check if all necessary parameters are given
	if(hostfile.empty())
	{
		cout << "Missing the non-optional option -h!" << endl;
		return 0;
	}
	if(guestfile.empty())
	{
		cout << "Missing the non-optional option -g!" << endl;
		return 0;
	}
	if(outputfile.empty())
		outputfile = "output.tga";

	//retrieve file size information
	int inputSize = getFileSize(hostfile);
	int hideSize = getFileSize(guestfile);

	//Open existing TGA
	ifstream tgaIn(hostfile.c_str(), ios::binary);
	if(tgaIn.fail())
	{
		cout << "Failed to open " << hostfile << endl;
		return 0;
	}

	//Open existing file that should be hidden
	ifstream hideIn(guestfile.c_str(), ios::binary);
	if(hideIn.fail())
	{
		cout << "Failed to open " << guestfile << endl;
		return 0;
	}

	//Check header of TGA file
	const int HEADERSIZE = 12;
	const char perfectHeader[HEADERSIZE] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	for(i = 0; i < HEADERSIZE; ++i)
	{
		char temp;
		if((temp = tgaIn.get()) != perfectHeader[i])
		{
			cout << "The file " << hostfile << " can not be used,\n"
				<< "the header (" << static_cast<unsigned int>(temp) << ") differs from " << static_cast<unsigned int>(perfectHeader[i]) << " at offset " << i << "." << endl;
			return 0;
		}
	}

	//Read additional header data
	unsigned short width, height;
	char colorDepth;
	tgaIn.read(reinterpret_cast<char *>(&width), sizeof(short));
	tgaIn.read(reinterpret_cast<char *>(&height), sizeof(short));
	tgaIn.read(&colorDepth, sizeof(char));
	if(colorDepth < 16)
	{
		cout << "Color depth ought to be at least 16!" << endl;
		return 0;
	}

	//Descriptor byte
	unsigned char dByte;
	tgaIn.read(reinterpret_cast<char *>(&dByte), sizeof(char));
	bool error = false;
	switch(colorDepth)
	{
	case 16:
		if((dByte & 0x0F) != 0x01)
			error = true;
		break;
	case 24:
		if((dByte & 0x0F) != 0x00)
			error = true;
		break;
	case 32:
		if((dByte & 0x0F) != 0x08)
			error = true;
		break;
	default:
		error = true;
		break;
	}
	if(dByte & 0xF0)
		error = true;
	if(error)
	{
		cout << "Error in image descriptor byte!\nHostfile is not suitable!" << endl;
		return 0;
	}

	//Do stuff for 16 bit images
	if(colorDepth == 16)
	{
		if(bitsperbyte == 4)
		{
			cout << "Color depth 16 bit is incompatible with 4 bits per byte.\n"
				"Truncating bits per byte to 2." << endl;
			bitsperbyte = 2;
		}
		if(bitsperbyte == 2)
		{
			transferByte = transfer16Byte2;
			bytesperbyte = 4;
		}
		if(bitsperbyte == 1)
		{
			transferByte = transfer16Byte1;
			bytesperbyte = 6;
		}
	}

	//We need to save the filename (stringlength + 1), the filesize (4) and the actual data (outputSize)
	int neededBytes = bytesperbyte * (static_cast<int>(guestfile.length()) + hideSize + 4 + 1);
	int availibleBytes = width * height * (colorDepth >> 3);
	//Check if image is big enough
	if(neededBytes > availibleBytes)
	{
		cout << "Hostfile is too small by " << neededBytes - availibleBytes << "!" << endl;
		return 0;
	}

	cout << "Hostfile was checked and approved.\nBeginning to hide " << guestfile << " with " << hideSize << " bytes." << endl;

	//Open output TGA file
	ofstream tgaOut(outputfile.c_str(), ios::binary);

	//Write header data to output TGA file
	for(i = 0; i < HEADERSIZE; ++i)
		tgaOut.write(&perfectHeader[i], 1);
	tgaOut.write(reinterpret_cast<char *>(&width), sizeof(width));
	tgaOut.write(reinterpret_cast<char *>(&height), sizeof(height));
	tgaOut.write(&colorDepth, 1);
	tgaOut.write(reinterpret_cast<char *>(&dByte), 1);

	//Perform actual hiding operation
	int filePos = 0;
	//Hide filename as 0 terminated string
	for(i = 0; i < int(guestfile.length()); ++i, filePos += bytesperbyte)
		transferByte(tgaIn, tgaOut, static_cast<unsigned char>(guestfile[i]), encryption);
	transferByte(tgaIn, tgaOut, 0, encryption);
	filePos += bytesperbyte;
	//Hide hidden file size
	for(i = 0; i < 4; ++i, filePos += bytesperbyte)
		transferByte(tgaIn, tgaOut, (reinterpret_cast<char *>(&hideSize))[i], encryption);
	//Hide the file
	for(i = 0; i < hideSize; ++i, filePos += bytesperbyte)
	{
		char tmp;
		hideIn.read(&tmp, 1);
		transferByte(tgaIn, tgaOut, static_cast<unsigned char>(tmp), encryption);
	}
	hideIn.close();
	if(encryption)
		delete encryption;
	//Transfer the rest of the file
	if(filePos < availibleBytes)
	{
		cout << "Beginning to copy " << availibleBytes - filePos << " unused bytes." << endl;
		if(scramble)
		{
			cout << "Scrambling up the unused bytes due to -s,\nthis might be very slow." << endl;
			for(i = 0; i < availibleBytes - filePos; ++i)
				transferByte(tgaIn, tgaOut, rand() & 0xFF, 0);
		}
		else
		{
			char *tmp = new char[availibleBytes - filePos];
			tgaIn.read(tmp, availibleBytes - filePos);
			tgaOut.write(tmp, availibleBytes - filePos);
			delete [] tmp;
		}
	}
	cout << "File was successfully hidden." << endl;
	return 0;
}
