//Stegano TGA Retrieve - A Polecatsoft program
//The program will retrieve a file which was hidden using Stegano TGA,
//in the file specified in parameter 1.

#include <string>
#include <iostream>
#include <fstream>

#include "main.h"
#include "XorEncryption.h"

using namespace std;

//For 16 bit images:
//This function returns a byte extracted from 6 bytes read from input
unsigned char retrieve16Byte1(ifstream &input, encryptClass *encryption)
{
	unsigned char retVal = 0;
	unsigned short readData;
	//Get the first 6 bits
	for(int i = 0; i < 6; i += 3)
	{
		input.read(reinterpret_cast<char*>(&readData), 2);

		retVal |=
			  ((readData & 0x0400) >> (10 - i))
			| ((readData & 0x0020) >> (4 - i)) // == 5 - (i + 1)
			| ((readData & 0x0001) << (i + 2));
	}
	//Get the last 2 bits
	input.read(reinterpret_cast<char*>(&readData), 2);

	if(encryption)
	{
		retVal |= ((readData & 0x0400) >> 4) | ((readData & 0x0020) << 2);
		encryption->exe(retVal);
		return retVal;
	}
	else
		return (retVal | ((readData & 0x0400) >> 4) | ((readData & 0x0020) << 2));
}

//This function returns a byte extracted from 4 bytes read from input
unsigned char retrieve16Byte2(ifstream &input, encryptClass *encryption)
{
	//Get the first 6 bits
	unsigned short readData;
	input.read(reinterpret_cast<char*>(&readData), 2);

	unsigned char retVal = ((readData & 0x0C00) >> 10)
		| ((readData & 0x0060) >> 3)
		| ((readData & 0x0003) << 4);

	//Get the last 2 bits
	input.read(reinterpret_cast<char*>(&readData), 2);

	if(encryption)
	{
		retVal |= ((readData & 0x0400) >> 4) | ((readData & 0x0020) << 2);
		encryption->exe(retVal);
		return retVal;
	}
	else
		return retVal | ((readData & 0x0400) >> 4) | ((readData & 0x0020) << 2);
}

//For 24 and 32 bit images:
//This function returns a byte extracted from 8 bytes read from input
unsigned char retrieveByte1(ifstream &input, encryptClass *encryption)
{
	unsigned char retVal = 0;
	unsigned char readByte;
	for(int i = 0; i < 8; ++i)
	{
		//Get value of active bit
		input.read(reinterpret_cast<char *>(&readByte), 1);
		retVal |= (readByte & 0x01) << i;
	}
	if(encryption)
		encryption->exe(retVal);
	return retVal;
}

//This function returns a byte extracted from 4 bytes read from input
unsigned char retrieveByte2(ifstream &input, encryptClass *encryption)
{
	unsigned char retVal = 0;
	unsigned char readByte;
	for(int i = 0; i < 8; i += 2)
	{
		input.read(reinterpret_cast<char *>(&readByte), 1);
		retVal |= (readByte & 0x03) << i;
	}
	if(encryption)
		encryption->exe(retVal);
	return retVal;
}

//This function returns a byte extracted from 2 bytes read from input
unsigned char retrieveByte4(ifstream &input, encryptClass *encryption)
{
	unsigned char retVal = 0;
	unsigned char readByte;
	for(int i = 0; i < 8; i += 4)
	{
		input.read(reinterpret_cast<char *>(&readByte), 1);
		retVal |= (readByte & 0x0F) << i;
	}
	if(encryption)
		encryption->exe(retVal);
	return retVal;
}

int steganoTgaRetrieve(int argc, char **argv)
{
	//for() counter variable
	int i;

	//Check if enough parameters are given
	cout << "You are now running the Stegano TGA Retrieve function of MLSS.\n" << endl;
	if(argc < 3)
	{
		cout << "Possible options (explanations are in brackets):\n"
			<< "-h hostfile (TGA file which contains hidden data)\n"
			<< "-p password (password will be used to decrypt the file)\n"
			<< "-b bitsperbyte (defines how many bits of each byte of the hostfile will be used to hide data. Possible values are 1, 2 & 4 (only 1 & 2 for 16 bit images).\nDefault: 1)\n";

		cout << "Non-optional options: -h" << endl;
		return 0;
	}

	//File names
	string hostfile;
	//How many bits of each hostfile byte will be used to hide data
	int bitsperbyte = 1;
	encryptClass *encryption = 0;
	unsigned char (*retrieveByte)(ifstream &input, encryptClass *encryption) = retrieveByte1;
	for(i = 2; i < argc; ++i)
	{
		string temp(argv[i]);
		bool missingArgument = false;
		if(temp == string("-h"))
		{
			if(++i < argc)
			{
				hostfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-b"))
		{
			if(++i < argc)
			{
				bitsperbyte = str2int(argv[i]);
				if(bitsperbyte == 1)
					retrieveByte = retrieveByte1;
				else if(bitsperbyte == 2)
					retrieveByte = retrieveByte2;
				else if(bitsperbyte == 4)
					retrieveByte = retrieveByte4;
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-p"))
		{
			if(++i < argc)
			{
				string password = argv[i];
				encryption = new encryptClass(password);
				continue;
			}
			missingArgument = true;
		}
		if(missingArgument)
			cout << "The option " << temp << " is missing an argument!";
		else
			cout << "Unknown option \"" << temp << "\"!";
		cout << endl;
		return 0;
	}

	//hostfile
	if(hostfile.empty())
	{
		cout << "Missing the non-optional option -h!" << endl;
		return 0;
	}

	//Open existing TGA
	ifstream tgaIn(hostfile.c_str(), ios::binary);
	if(tgaIn.fail())
	{
		cout << "Failed to open " << hostfile << endl;
		return 0;
	}

	//Check header of TGA file
	const int HEADERSIZE = 12;
	char perfectHeader[HEADERSIZE] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	for(i = 0; i < HEADERSIZE; ++i)
	{
		char temp;
		if((temp = tgaIn.get()) != perfectHeader[i])
		{
			cout << "The file " << hostfile << " can not be used,\n"
				<< "the header (" << static_cast<unsigned int>(temp) << ") differs from " << static_cast<unsigned int>(perfectHeader[i]) << " at offset " << i << "." << endl;
			return 0;
		}
	}

	//Read additional header data
	unsigned short width, height;
	char colorDepth;
	tgaIn.read(reinterpret_cast<char *>(&width), sizeof(short));
	tgaIn.read(reinterpret_cast<char *>(&height), sizeof(short));
	tgaIn.read(&colorDepth, sizeof(char));
	if(colorDepth < 16)
	{
		cout << "Color depth ought to be at least 16!" << endl;
		return 0;
	}

	//Descriptor byte
	unsigned char dByte;
	tgaIn.read(reinterpret_cast<char *>(&dByte), sizeof(char));
	bool error = false;
	switch(colorDepth)
	{
	case 16:
		if((dByte & 0x0F) != 0x01)
			error = true;
		break;
	case 24:
		if((dByte & 0x0F) != 0x00)
			error = true;
		break;
	case 32:
		if((dByte & 0x0F) != 0x08)
			error = true;
		break;
	default:
		error = true;
		break;
	}
	if(dByte & 0xF0)
		error = true;
	if(error)
	{
		cout << "Error in image descriptor byte!\nHostfile is not suitable!" << endl;
		return 0;
	}
 
	//Do stuff for 16 bit images
	if(colorDepth == 16)
	{
		if(bitsperbyte == 4)
		{
			cout << "Color depth 16 bit is incompatible with 4 bits per byte.\n"
				"Truncating bits per byte to 2." << endl;
			bitsperbyte = 2;
		}
		if(bitsperbyte == 2)
			retrieveByte = retrieve16Byte2;
		else if(bitsperbyte == 1)
			retrieveByte = retrieve16Byte1;
	}

	cout << "Input TGA file was checked and approved.\nBeginning to retrieve data." << endl;

	//Retrieve filename
	string filename;
	char readByte;
	while((readByte = retrieveByte(tgaIn, encryption)) != 0)
		filename += readByte;
	//Retrieve filesize
	int fileSize;
	for(i = 0; i < 4; ++i)
		(reinterpret_cast<char *>(&fileSize))[i] = retrieveByte(tgaIn, encryption);
	cout << "Creating the file " << filename << " with " << fileSize << " bytes." << endl;
	cout << endl;

	//Retrieve file content
	ofstream retrovenFile(filename.c_str(), ios::binary);
	if(retrovenFile.fail())
	{
		cout << "Failed to open " << filename << ", maybe no data is actually hidden in the image,\n"
			"or you supplied a wrong value for -b!" << endl;
		return 0;
	}
	readByte = retrieveByte(tgaIn, encryption);
	for(i = 0; i < fileSize; ++i, readByte = retrieveByte(tgaIn, encryption))
		retrovenFile.write(&readByte, 1);
	retrovenFile.close();
	tgaIn.close();

	if(encryption)
		delete encryption;
	cout << "File was successfully retrieved." << endl;
	return 0;
}
