//Master Lenman's Text Compression and Decompression - A Polecatsoft program
//This program uses the MLTC Algorithm class to compress and decompress text files

#include <string>
#include <iostream>
#include <fstream>

#include "mltc.h"
#include "main.h"
#include "treeNode.h"
#include "treeStringClass.h"
#include "MLTCClass.h"

using namespace std;

//This will compress a file
int mltcompress(int argc, char **argv)
{
	cout << "You are now running Master Lenman's Text Compression.\n" << endl;
	//Show help if neccessary
	if(argc < 3)
	{
		cout << "Possible options (explanations are in brackets):\n"
			<< "-i inputfile (the file that will be compressed)\n"
			<< "-o outputfile (the compressed data will be written in here)\n";

		cout << "Non-optional options: -i -o" << endl;
		return 0;
	}
	//Parse parameters
	string inputfile, outputfile;
	for(int i = 2; i < argc; ++i)
	{
		string temp(argv[i]);
		bool missingArgument = false;
		if(temp == string("-i"))
		{
			if(++i < argc)
			{
				inputfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-o"))
		{
			if(++i < argc)
			{
				outputfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		if(missingArgument)
			cout << "The option " << temp << " is missing an argument!";
		else
			cout << "Unknown option \"" << temp << "\"!";
		cout << endl;
		return 0;
	}

	//Check if all necessary parameters are given
	if(inputfile.empty())
	{
		cout << "Missing the non-optional option -i!" << endl;
		return 0;
	}
	if(outputfile.empty())
	{
		cout << "Missing the non-optional option -o!" << endl;
		return 0;
	}

	//Read in the input file and put it into the compression stream
	ifstream input(inputfile.c_str(), ios::binary);
	if(input.fail())
	{
		cout << "Failed to open \"" << inputfile << "\"!" << endl;
		return 0;
	}
	cout << "Feeding input file into compression stream..." << flush;
	unsigned int fileSize = getFileSize(inputfile), fileSizeLeft = fileSize;

	const int STEPSIZE = fileSize / 100;
	int stepCount = STEPSIZE;
	const char BACKSPACE = char(0x08);
	int percentage = 0;
	cout << "00%";

	char *readBuffer = new char[1024];
	MLTCStream stream;
	int useValue = 1024;
	while(fileSizeLeft)
	{
		if(fileSizeLeft < 1024)
		{
			useValue = fileSizeLeft;
		}
		input.read(readBuffer, useValue);
		stream.write(readBuffer, useValue);
		fileSizeLeft -= useValue;

		if((stepCount -= useValue) <= 0)
		{
			cout << BACKSPACE << BACKSPACE << BACKSPACE;
			if(++percentage < 10)
				cout << "0";
			cout << percentage << "%";
			stepCount += STEPSIZE;
		}
	}
	input.close();
	delete [] readBuffer;
	cout << BACKSPACE << BACKSPACE << BACKSPACE << BACKSPACE << "Done\nCompressing..." << flush;
	//Compress
	stream.compress(true);
	
	//Put compressed data into the output file
	cout << "Done\nWriting compressed data to output file..." << flush;
	ofstream output(outputfile.c_str(), ios::binary);
	output.write(stream.getOutputBuffer(), stream.getOutputSize());
	output.close();
	cout << "Done\n" << endl;
	cout << "Compression ratio: " << float(stream.getOutputSize()) / float(fileSize) << endl;
	cout << "Average bits per byte: " << stream.getBitsPerByte() << endl;
	return 0;
}

//This will decompress a file
int mltdecompress(int argc, char **argv)
{
	cout << "You are now running Master Lenman's Text Decompression.\n" << endl;
	//Show help if neccessary
	if(argc < 3)
	{
		cout << "Possible options (explanations are in brackets):\n"
			<< "-i inputfile (the file that contains compressed data)\n"
			<< "-o outputfile (the decompressed data will be written here)\n";

		cout << "Non-optional options: -i -o" << endl;
		return 0;
	}
	//Parse parameters
	string inputfile, outputfile;
	for(int i = 2; i < argc; ++i)
	{
		string temp(argv[i]);
		bool missingArgument = false;
		if(temp == string("-i"))
		{
			if(++i < argc)
			{
				inputfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-o"))
		{
			if(++i < argc)
			{
				outputfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		if(missingArgument)
			cout << "The option " << temp << " is missing an argument!";
		else
			cout << "Unknown option \"" << temp << "\"!";
		cout << endl;
		return 0;
	}

	//Check if all necessary parameters are given
	if(inputfile.empty())
	{
		cout << "Missing the non-optional option -i!" << endl;
		return 0;
	}
	if(outputfile.empty())
	{
		cout << "Missing the non-optional option -o!" << endl;
		return 0;
	}

	//Read in the input file and put it into the compression stream
	ifstream input(inputfile.c_str(), ios::binary);
	if(input.fail())
	{
		cout << "Failed to open \"" << inputfile << "\"!" << endl;
		return 0;
	}
	cout << "Feeding input file into decompression stream..." << flush;
	unsigned int fileSize = getFileSize(inputfile), fileSizeLeft = fileSize;

	const int STEPSIZE = fileSize / 100;
	int stepCount = STEPSIZE;
	const char BACKSPACE = char(0x08);
	int percentage = 0;
	cout << "00%";

	char *readBuffer = new char[1024];
	MLTCStream stream;
	int useValue = 1024;
	while(fileSizeLeft)
	{
		if(fileSizeLeft < 1024)
		{
			useValue = fileSizeLeft;
		}
		input.read(readBuffer, useValue);
		stream.write(readBuffer, useValue);
		fileSizeLeft -= useValue;

		if((stepCount -= useValue) <= 0)
		{
			cout << BACKSPACE << BACKSPACE << BACKSPACE;
			if(++percentage < 10)
				cout << "0";
			cout << percentage << "%";
			stepCount += STEPSIZE;
		}
	}
	input.close();
	delete [] readBuffer;
	cout << BACKSPACE << BACKSPACE << BACKSPACE << BACKSPACE << "Done\nDecompressing..." << flush;
	//Decompress
	stream.decompress(true);
	
	//Put decompressed data into the output file
	cout << "Done\nWriting decompressed data to output file..." << flush;
	ofstream output(outputfile.c_str(), ios::binary);
	output.write(stream.getOutputBuffer(), stream.getOutputSize());
	output.close();
	cout << "Done\n" << endl;
	return 0;
}
